%% Lecture 6 - video and image processing
% Sampling frequency - 1000Hz
% Image resolution: The triangular tracking marker consisted of three black
% points representing the vertices of an isosceles triangle (measuring 2.7  2.3 mm;

close all; clear all; clc; format compact;

recompute_import = 'false';
recompute_track = 'false';
recompute_scaling = 'true';

if ~exist('BeeFlight_data.mat', 'file') || strcmp(recompute_import, 'true')
    disp('loading video')
    %% Load and display information
    %Create a video object from the file
    v = VideoReader('BeeFlight.mov');
    
    %Display information about the video
    get(v)
    
    %% Import and preprocess data
    
    %Preallocate
    N = v.NumberOfFrames; %number of frames
    N1 = v.Height; %dimensions of frames
    N2 = v.Width;
    
    imGray = zeros(N1, N2, N, 'uint8');
    
    %Import and convert to grayscale image stack
    for k = 1:N
        imGray(:,:,k) = rgb2gray(read(v, k));
    end
    
    %Threshold and store BW image stack
    imBW = false(N1, N2, N); %Preallocate array 
    threshold = 0.5;    %Try an initial threshold 
    for iIm = 1:N 
        disp(['loading video: ', num2str(iIm), '/', num2str(N)])
        imBW(:,:,iIm) = im2bw(imGray(:,:,iIm), threshold);
        if iIm == 1
            while true
                figure(1)
                subplot(1,2,1)
                imshow(imGray(:,:,iIm))
                subplot(1,2,2)
                imshow(imBW(:,:,iIm));
                str = input('Acceptable threshold? (y/n)','s');
                if strcmp(str, 'y')
                    break
                elseif strcmp(str, 'n')
                    threshold = input('Please enter new threshold: ');
                    imBW(:,:,iIm) = im2bw(imGray(:,:,iIm), threshold);
                else
                    disp('Input not recognized')
                end
            end
            close 1
        end
    end
    
    
    
    %Save image stack
    disp('saving .mat file')
    save('BeeFlight_data.mat', 'imGray', 'imBW', 'N', 'N1', 'N2')
else
    load('BeeFlight_data.mat')
end

%% Find bee head in each frame

if ~exist('BeeFlight_track.mat', 'file') || strcmp(recompute_track, 'true')
    
    data = struct('time', [], 'centroid', []);
    freq = 1000; %sampling frequency, [Hz]
    
    
    %Invert images; Matlab by default finds white objects in black
    %backgrounds
    imBW = ~imBW;
    
    
    for iIm = 1:N;
        disp(['processing video: ', num2str(iIm),'/',num2str(N)])
        data(iIm).time = (iIm-1)/freq; %[s]
        
        %Clean up the image a bit
        im = bwmorph(imBW(:,:,iIm), 'majority','4');
        im = bwmorph(im,'clean', '4');
        im = bwmorph(im,'spur', '4');
        
        %Find the boundaries & pixel indices in each region
        [B,~,Nb] = bwboundaries(im);
        stats = regionprops(im,'PixelList', 'Centroid');
        
        if iIm == 1 %If this is the first image, ask user to identify bee head
            
            %Plot original image
            figure(1)
            imshow(imGray(:,:,iIm))
            hold on
            
            %Ask user to click on bee head
            disp('click on target');
            [xB, yB] = ginput(1);
            
            xB = round(xB);
            yB = round(yB);
            
            % Identify which boundary contains this point and plot
            for k = 1:Nb
                if ismember([xB,yB],stats(k).PixelList, 'rows')
                    boundary = B{k};
                    plot(boundary(:,2),boundary(:,1),'r','LineWidth',1.5);
                    data(iIm).centroid = stats(k).Centroid;
                    break
                else
                    if k==Nb
                        error(sprintf(['Unable to find region containing ',...
                            'selected point.\nPlease either rerun program ',...
                            'selecting alternate point or alter image ',...
                            'thresholding/preprocessing.']))
                    end
                end
            end
            pause(1)
            
        else %find the region with the closest centroid to the prior image
            if ~isempty(stats)
                centroids = cat(1, stats.Centroid);
                idx = knnsearch(centroids, data(iIm-1).centroid);
                data(iIm).centroid = stats(idx).Centroid;
            else
                data(iIm).centroid = [NaN, NaN];
            end
        end
        
        
        
    end
    
    save('BeeFlight_track.mat', 'data', 'N')
else
    load('BeeFlight_track.mat')
end


%% Make a movie of results
disp('make movie of results')
warning off

M(N) = struct('cdata',[],'colormap',[]);
path = [];

for iIm = 1:N;
    disp(['tracking bee movie: ', num2str(iIm), '/', num2str(N)])
    path = cat(1, path, data(iIm).centroid);
    
    figure(1)
    imshow(imGray(:,:,iIm))
    hold on
    
    plot(path(:,1), path(:,2),'g-')
    plot(data(iIm).centroid(1), data(iIm).centroid(2),'r.');
   
    text(162, 50,sprintf('Time: %4.3f seconds', ...
        data(iIm).time), 'FontSize', 20, 'Color', 'w')
    
    M(iIm) = getframe(gcf);
    clf
end

movie2avi(M, 'BeeFlight_headPath.avi', 'compression', 'none', 'fps', 60)



%% %Extract scaling factor
disp('extract scaling factor')
if ~exist('BeeFlight_scaling.mat', 'file') || strcmp(recompute_scaling, 'true')
    
    figure(1)
    imshow(imGray(:,:,500))
    hold on
    xlim([304.0519  431.4687])
    ylim([ 398.4902  525.9070])
    [xT, yT] = ginput(3);
    plot(xT, yT, 'r.')
    
    l1 = sqrt((xT(1)-xT(2)).^2 + (yT(1)-yT(2)).^2);
    l2 = sqrt((xT(1)-xT(3)).^2 + (yT(1)-yT(3)).^2);
    l3 = sqrt((xT(2)-xT(3)).^2 + (yT(2)-yT(3)).^2);
    
    image_lengths = [l1, l2, l3];
    [~,idx] = min(image_lengths);
    physical_lengths = ones(1,3).*2.7;
    physical_lengths(idx) = 2.3;
    
    scale_mean = mean(physical_lengths./image_lengths); %[mm/pixel]
    scale_std = std(physical_lengths./image_lengths); %[mm/pixel]
    
    fprintf('The scaling factor is %3.2f +/- %3.2f mm/pixel.\n', scale_mean, scale_std);
    
    save('BeeFlight_scaling.mat', 'scale_mean', 'scale_std')
else
    load('BeeFlight_scaling.mat')
end

%% Compute bee velocity over time

time = [data.time];
position = cat(1,data.centroid);

%Differentials
dx = diff(position(:,1)).*scale_mean./1000; %[m]
dy = diff(position(:,2)).*scale_mean./1000; %[m]
ds = sqrt(dx.^2 + dy.^2);
dt = diff(time); %[s]

%Compute velocity 
velocity = ds./dt'; %[mm/s]

%Plot results
figure()
plot(time(2:end),velocity)
xlabel('Time, [s]')
ylabel('Speed (within image plane), [m/s]')


